#include <allegro.h>
#include <stdio.h>
#include <time.h>
#include "mode15k.h"
#include "s6326_15.h"
#include "g400__15.h"
#include "savage15.h"
#include "laguna15.h"
#include "tdfx__15.h"
#include "s3.h"



#define DEVICE_SIS6326	1
#define DEVICE_G400 	2
#define DEVICE_SAVAGE	3
#define DEVICE_3DFX 	4
#define DEVICE_LAGUNA	5
#define DEVICE_S3		6
#define DEVICE_BLADE	7


static int g_DeviceType = 0;


static FILE* g_scrmodelog = NULL;

static CRTCPARAM g_undo_cp1;
static CRTCPARAM g_undo_cp2;


int VESA_WIDTH	= 1024;
int VESA_HEIGHT = 768;
int VESA_BPP = 8;

#define INDEX_HAll		 0
#define INDEX_HSync 	 1
#define INDEX_HTotal	 2
#define INDEX_HDisp 	 3
#define INDEX_HSStart	 4
#define INDEX_HSEnd 	 5
#define INDEX_VAll		 6
#define INDEX_VSync 	 7
#define INDEX_VTotal	 8
#define INDEX_VDisp 	 9
#define INDEX_VSStart	 10
#define INDEX_VSEnd 	 11
#define INDEX_vert001Hz  12
#define INDEX_doublescan 13
#define INDEX_interlace  14
#define INDEX_hpolarity  15
#define INDEX_vpolarity  16

#define MAX_PARAM		 17

#define GET_320x240_60_15KHz   1
#define GET_512x384_60_24KHz   2
#define GET_640x480_60_31KHz   3
#define GET_512x384_60_31KHz   4
#define GET_F5	5
#define GET_F6	6
#define GET_F7	7
#define GET_F8	8
#define GET_F9	9
#define GET_F10 10
#define GET_F11 11
#define GET_F12 12




static long SiS6326_GetVclkParam(long vclk, int *Num, int *Div,
								 int *DeN, int *PS, int *SeqDiv)
{
	double desfreq;
	double temp;
	double tempdt;
	double dt;
	int d0;
	int d1;

	if(vclk > 175000000) vclk = 175000000;
	*SeqDiv = 1;
	*Div	= 1;
	*PS 	= 1;
	if( 	 vclk >= 150000000){
		*PS  = 1;
	}else if(vclk >= 125000000){
		*Div = 2;
		*PS  = 3;
	}else if(vclk >=  75000000){
		*PS  = 2;
	}else if(vclk >=  62500000){
		*PS  = 3;
	}else if(vclk >=  37500000){
		*PS  = 4;
	}else if(vclk >=  31250000){
		*PS  = 6;
	}else if(vclk >=  13500000){
		*PS  = 8;
	}else{
		*PS  = 8;
		*SeqDiv = 2;
	}

	desfreq = vclk * (*PS / (double)(*Div)) * (*SeqDiv);
	dt = 250000000.0;
	*Num = 0;
	*DeN = 0;
	d0 = 2;
	while(d0 < 128){
		d1 = 3;
		while(d1 < 32){
			temp   = 14318000 * (double)d0 / d1;
			tempdt = desfreq - temp;
			if(tempdt < 0) tempdt *= -1;
			if(tempdt < dt){
				*Num = d0;
				*DeN = d1;
				dt = tempdt;
			}
			d1 += 1;
		}
		d0 += 1;
	}

	vclk = 14318000 * (*Num * *Div)/(*DeN * *PS * *SeqDiv);

#if 0
	logerror("Num %3d, DeN %3d, Div %1d, PS %1d, SeqDiv %1d\n",
				*Num, *DeN, *Div, *PS, *SeqDiv);
	logerror("14.318 * (Num/DeN) : %lf\n", 14.318*(*Num)/(*DeN));
	logerror("vclk : %lf\n", (double)vclk / 1000000.0);
#endif

	return vclk;
}





void DispCP(CRTCPARAM* cp)
{
	if(g_scrmodelog == NULL) return;

	fprintf(g_scrmodelog, "\n");
	fprintf(g_scrmodelog, "\n");
	fprintf(g_scrmodelog, "\n");

	fprintf(g_scrmodelog, "static CRTCPARAM mode%dx%d_%d = { \n", cp->width, cp->height,
					(int)(cp->vert001Hz + 50) / 100
	);

	fprintf(g_scrmodelog, "// HTotal Disp BSta SSta SEnd BEnd VTotal Disp BSta SSta SEnd BEnd\n");

	fprintf(g_scrmodelog, "   ");
	fprintf(g_scrmodelog, "%6d,", cp->HTotal);
	fprintf(g_scrmodelog, "%4d,", cp->HDisp);
	fprintf(g_scrmodelog, "%4d,", cp->HBStart);
	fprintf(g_scrmodelog, "%4d,", cp->HSStart);
	fprintf(g_scrmodelog, "%4d,", cp->HSEnd);
	fprintf(g_scrmodelog, "%4d,", cp->HBEnd);
	fprintf(g_scrmodelog, "%6d,", cp->VTotal);
	fprintf(g_scrmodelog, "%4d,", cp->VDisp);
	fprintf(g_scrmodelog, "%4d,", cp->VBStart);
	fprintf(g_scrmodelog, "%4d,", cp->VSStart);
	fprintf(g_scrmodelog, "%4d,", cp->VSEnd);
	fprintf(g_scrmodelog, "%4d,", cp->VBEnd);

	fprintf(g_scrmodelog, "\n");

	fprintf(g_scrmodelog, "\n// width height HPlr VPlr DblScn Intrlc Iratio \n");
	fprintf(g_scrmodelog, "%8d,", cp->width);
	fprintf(g_scrmodelog, "%6d,", cp->height);
	fprintf(g_scrmodelog, "%4d, ", cp->hpolarity);
	fprintf(g_scrmodelog, "%3d, ", cp->vpolarity);
	fprintf(g_scrmodelog, "%5d, ", cp->doublescan);
	fprintf(g_scrmodelog, "%5d, ", cp->interlace);
	fprintf(g_scrmodelog, "%5d, ", cp->interlaceratio);

	fprintf(g_scrmodelog, "\n");

	fprintf(g_scrmodelog, "\n// dotclkHz  horzHz  vertHz(in 0.01)\n");
	fprintf(g_scrmodelog, "%11ld,", cp->dotclockHz);
	fprintf(g_scrmodelog, "%7ld,", cp->horzHz);
	fprintf(g_scrmodelog, "%7ld", cp->vert001Hz);
	fprintf(g_scrmodelog, "\n};\n");

	fflush(g_scrmodelog);
}














BOOL InitGfxMode(void)
{
	set_color_depth(VESA_BPP);
	if(set_gfx_mode(GFX_VESA2L, VESA_WIDTH, VESA_HEIGHT,
								 VESA_WIDTH, VESA_HEIGHT) != 0){
		if(set_gfx_mode(GFX_VESA1, VESA_WIDTH, VESA_HEIGHT,
									 VESA_WIDTH, VESA_HEIGHT) != 0){

			return FALSE;
		}
	}
	set_clip(screen, 0, 0, VESA_WIDTH-1, VESA_HEIGHT-1);

	return TRUE;
}

void ExitGfxMode(void)
{
	set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
}



void WriteGrid(int width, int height, int i,
			   int edge,  int edge_x, int edge_y,
			   int interlaceflag, int doublescan)
{
	int x;
	int y;

	if(i == 0)
		i = 2;
	else
		i = 1;

	if(interlaceflag != 0)
		edge_y = ((edge_y + 1) * 2) - 1;
	if(doublescan != 0)
		edge_y = ((edge_y + 1) / 2) - 1;

	clear(screen);

	// 16x16 blue
	y = 0;
	while(y < height){
		if((y % 16) == 0){
			hline(screen, 0, y, width - 1, makecol(0, 0, 96/i));
		}
		y += 1;
	}
	x = 0;
	while(x < width){
		if((x % 16) == 0){
			vline(screen, x, 0, height - 1, makecol(0, 0, 96/i));
		}
		x += 1;
	}

	// 32x32 white
	y = 0;
	while(y < height){
		if((y % 32) == 0){
			hline(screen, 0, y, width - 1, makecol(255/i, 255/i, 255/i));
		}
		y += 1;
	}
	x = 0;
	while(x < width){
		if((x % 32) == 0){
			vline(screen, x, 0, height - 1, makecol(255/i, 255/i, 255/i));
		}
		x += 1;
	}

	// 128x128 green
	y = 0;
	while(y < height){
		if((y % 128) == 0){
			hline(screen, 0, y, width - 1, makecol(0, 255/i, 0));
		}
		y += 1;
	}
	x = 0;
	while(x < width){
		if((x % 128) == 0){
			vline(screen, x, 0, height - 1, makecol(0, 255/i, 0));
		}
		x += 1;
	}

	if(edge != 0){
			hline(screen, 0,	  0, width - 1,  makecol(0, 255/i, 255/i));
			hline(screen, 0, edge_y, width - 1,  makecol(0, 255/i, 255/i));
			vline(screen, edge_x, 0, height - 1, makecol(0, 255/i, 255/i));
			vline(screen,	   0, 0, height - 1, makecol(0, 255/i, 255/i));
	}
}

void DispText(CRTCPARAM *cp, int cursorpos)
{
	int ox = 32;
	int oy = 16;
	int clm;
	int coloridx[MAX_PARAM];
	int d0;
	long temp;
	long temp2;
	long dotclock;

	d0 = 0;
	while(d0 < MAX_PARAM){
		coloridx[d0] = makecol(255, 255, 255);
		d0 += 1;
	}
	if(cursorpos < 0)  cursorpos = 0;
	if(cursorpos > (MAX_PARAM - 1)) cursorpos = MAX_PARAM - 1;
	coloridx[cursorpos] = makecol(255, 255, 0);

	d0 = 0;
	clm = 0;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"Horz ALL");   clm++;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"Horz Sync");  clm++;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"HTotal%6d",  cp->HTotal);	clm++;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"HDisp%7d",   cp->HDisp);	clm++;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"HSStart%5d", cp->HSStart); clm++;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"HSEnd%7d",   cp->HSEnd);	clm++;
	clm++;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"Vert ALL");   clm++;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"Vert Sync");  clm++;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"VTotal%6d",  cp->VTotal);	clm++;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"VDisp%7d",   cp->VDisp);	clm++;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"VSStart%5d", cp->VSStart); clm++;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"VSEnd%7d",   cp->VSEnd);	clm++;
	clm++;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"Vert%5d.%02dHz",  cp->vert001Hz/100, cp->vert001Hz % 100);  clm++;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"DoubleScan%2d",   cp->doublescan);   clm++;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"Interlace%3d", cp->interlace);   clm++;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"HPolarity%3d", cp->hpolarity);   clm++;
	textprintf(screen, font, ox, oy + clm * text_height(font), coloridx[d0++],
			"VPolarity%3d", cp->vpolarity);   clm++;

	clm++;

	temp = cp->horzHz;
	textprintf(screen, font, ox, oy + clm * text_height(font), makecol(255, 255, 255),
			"Horz%5ld.%02ldKHz",  temp / 1000, (temp/10) % 100);  clm++;

	dotclock = cp->dotclockHz;
	temp = dotclock;
	textprintf(screen, font, ox, oy + clm * text_height(font), makecol(255, 255, 255),
			"DClk%5ld.%02ldMHz", temp/1000000, (temp/10000) % 100); clm++;
	{
		int Num, Div, DeN, PS, SeqDiv;

		temp2 = SiS6326_GetVclkParam(dotclock, &Num, &Div, &DeN, &PS, &SeqDiv);
	}
	clm++;

	if(g_DeviceType == DEVICE_SIS6326){
		temp = ((double)temp2 * 100) / (cp->HTotal * cp->VTotal);
		textprintf(screen, font, ox, oy + clm * text_height(font), makecol(255, 255, 255),
				"SiS6326 Vert%4d.%02dHz",  temp/100, temp % 100);  clm++;
		temp = temp2 / cp->HTotal;
		textprintf(screen, font, ox, oy + clm * text_height(font), makecol(255, 255, 255),
				"SiS6326 Horz%4ld.%02ldKHz",  temp / 1000, (temp/10) % 100);  clm++;
	}
}

void AdjustCRTCPARAM(CRTCPARAM *cp)
{
	double VTotal;
	int Num, Div, DeN, PS, SeqDiv;

	if(cp->interlace == 0){
		VTotal = cp->VTotal;
	}else{
		VTotal = cp->VTotal + 0.5;
	}

	cp->HBStart = cp->HDisp;
	cp->HBEnd	= cp->HTotal - 8;
	cp->VBStart = cp->VDisp;
	cp->VBEnd	= cp->VTotal - 1;

	cp->horzHz = (VTotal * cp->vert001Hz) / 100;
	cp->dotclockHz = (long)((double)VTotal * cp->HTotal * cp->vert001Hz / 100.0);

	cp->width  = cp->HDisp;
	if(cp->interlace != 0){
		cp->height = cp->VDisp * 2;
	}else if(cp->doublescan != 0){
		cp->height = cp->VDisp / 2;
	}else{
		cp->height = cp->VDisp;
	}

	cp->interlaceratio = 0;
}


void IncParam(int value, CRTCPARAM* cp, int cursorpos)
{
	int d0;

	if(cursorpos < 0) cursorpos = 0;
	if(cursorpos > (MAX_PARAM - 1)) cursorpos = MAX_PARAM - 1;

	switch(cursorpos){
		case INDEX_HAll:
			value *= 8;
			d0 = cp->HTotal + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->HTotal = d0;

			d0 = cp->HDisp + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->HDisp = d0;

			d0 = cp->HSStart + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->HSStart = d0;

			d0 = cp->HSEnd + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->HSEnd = d0;
			break;

		case INDEX_VAll:
			d0 = cp->VTotal + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->VTotal = d0;

			d0 = cp->VDisp + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->VDisp = d0;

			d0 = cp->VSStart + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->VSStart = d0;

			d0 = cp->VSEnd + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->VSEnd = d0;
			break;

		case INDEX_HSync:
			value *= 8;
			d0 = cp->HSStart + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->HSStart = d0;

			d0 = cp->HSEnd + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->HSEnd = d0;
			break;

		case INDEX_VSync:
			d0 = cp->VSStart + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->VSStart = d0;

			d0 = cp->VSEnd + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->VSEnd = d0;
			break;

		case INDEX_HTotal:
			value *= 8;
			d0 = cp->HTotal + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->HTotal = d0;
			break;

		case INDEX_HDisp:
			value *= 8;
			d0 = cp->HDisp + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->HDisp = d0;
			break;

		case INDEX_HSStart:
			value *= 8;
			d0 = cp->HSStart + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->HSStart = d0;
			break;

		case INDEX_HSEnd:
			value *= 8;
			d0 = cp->HSEnd + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->HSEnd = d0;
			break;


		case INDEX_VTotal:
			d0 = cp->VTotal + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->VTotal = d0;
			break;

		case INDEX_VDisp:
			d0 = cp->VDisp + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->VDisp = d0;
			break;

		case INDEX_VSStart:
			d0 = cp->VSStart + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->VSStart = d0;
			break;

		case INDEX_VSEnd:
			d0 = cp->VSEnd + value;
			if(d0 < 8)	  d0 = 8;
			if(d0 > 2048) d0 = 2048;
			cp->VSEnd = d0;
			break;

		case INDEX_vert001Hz:
			cp->vert001Hz += value;
			if(cp->vert001Hz < 1000)  cp->vert001Hz = 1000;
			if(cp->vert001Hz > 16000) cp->vert001Hz = 16000;
			break;

		case INDEX_doublescan:
			cp->doublescan += value;
			if(cp->doublescan < 0) cp->doublescan = 0;
			if(cp->doublescan > 1) cp->doublescan = 1;
			if(cp->doublescan == 1)
				cp->interlace = 0;
			break;

		case INDEX_interlace:
			cp->interlace += value;
			if(cp->interlace < 0) cp->interlace = 0;
			if(cp->interlace > 1) cp->interlace = 1;
			if(cp->interlace == 1)
				cp->doublescan = 0;
			break;

		case INDEX_hpolarity:
			cp->hpolarity += value;
			if(cp->hpolarity < 0) cp->hpolarity = 0;
			if(cp->hpolarity > 1) cp->hpolarity = 1;
			break;

		case INDEX_vpolarity:
			cp->vpolarity += value;
			if(cp->vpolarity < 0) cp->vpolarity = 0;
			if(cp->vpolarity > 1) cp->vpolarity = 1;
			break;
	}

	if(cp->HDisp > VESA_WIDTH)	cp->HDisp = VESA_WIDTH;
	if(cp->VDisp > VESA_HEIGHT) cp->VDisp = VESA_HEIGHT;

}





static void SetMode15K(CRTCPARAM *cp)
{
	g_undo_cp2 = g_undo_cp1;
	g_undo_cp1 = *cp;
	SetCRTCPARAM(cp);
	switch(g_DeviceType){
		case DEVICE_SIS6326:
			setSiS6326();
			break;

		case DEVICE_G400:
			setG400();
			break;

		case DEVICE_SAVAGE:
			setSavage();
			break;

		case DEVICE_3DFX:
			set3dfx();
			break;

		case DEVICE_LAGUNA:
			setLaguna();
			break;

		case DEVICE_S3:
			setS3();
			break;

		case DEVICE_BLADE:
			setBlade();
			break;
	}
}


static BOOL DetectSVGA15KHz(void)
{
	if(detectSiS6326() != 0){
		g_DeviceType = DEVICE_SIS6326;
		return TRUE;
	}else if(detectG400() != 0){
		g_DeviceType = DEVICE_G400;
		return TRUE;
	}else if(detectSavage() != 0){
		g_DeviceType = DEVICE_SAVAGE;
		return TRUE;
	}else if(detect3dfx() != 0){
		g_DeviceType = DEVICE_3DFX;
		return TRUE;
	}else if(detectLaguna() != 0){
		g_DeviceType = DEVICE_LAGUNA;
		return TRUE;
	}else if(detectS3() != 0){
		g_DeviceType = DEVICE_S3;
		return TRUE;
	}else if(detectBlade() != 0){
		g_DeviceType = DEVICE_BLADE;
		return TRUE;
	}
	return FALSE;
}

static void ExitSVGA15KHz(void)
{
	switch(g_DeviceType){
		case DEVICE_SIS6326:
			resetSiS6326_15KHz();
			break;

		case DEVICE_G400:
			resetG400_15KHz();
			break;

		case DEVICE_SAVAGE:
			resetSavage_15KHz();
			break;

		case DEVICE_3DFX:
			reset3dfx_15KHz();
			break;

		case DEVICE_LAGUNA:
			resetLaguna_15KHz();
			break;

		case DEVICE_S3:
			resetS3();
			break;

		case DEVICE_BLADE:
			resetBlade_15KHz();
			break;
	}
}






int main(int argc, char* argv[])
{
	CRTCPARAM cp;
	int cursorpos;
	int keydata;
	int index;
	int grid_intensity = 0;
	int draw_edge_line = 0;
	int firstscrmode;

	firstscrmode = GET_640x480_60_31KHz;
	index = 1;
	while(index < argc){
		if( strcmp(argv[index], "-640x480@31") == 0 ){
			firstscrmode = GET_640x480_60_31KHz;
		}else if( strcmp(argv[index], "-320x240@15") == 0 ){
			firstscrmode = GET_320x240_60_15KHz;
		}else if( strcmp(argv[index], "-512x384@24") == 0 ){
			firstscrmode = GET_512x384_60_24KHz;
		}else if( strcmp(argv[index], "-512x384@31") == 0 ){
			firstscrmode = GET_512x384_60_31KHz;
		}else if( strcmp(argv[index], "-vesa640x480") == 0 ){
			VESA_WIDTH	= 640;
			VESA_HEIGHT = 480;
		}else if( strcmp(argv[index], "-vesa800x600") == 0 ){
			VESA_WIDTH	= 800;
			VESA_HEIGHT = 600;
		}else if( strcmp(argv[index], "-vesa1024x768") == 0 ){
			VESA_WIDTH	= 1024;
			VESA_HEIGHT = 768;
		}else if( strcmp(argv[index], "-vesa1280x1024") == 0 ){
			VESA_WIDTH	= 1280;
			VESA_HEIGHT = 1024;
		}else if( strcmp(argv[index], "-vesa1600x1200") == 0 ){
			VESA_WIDTH	= 1600;
			VESA_HEIGHT = 1200;
		}else if( strcmp(argv[index], "-16bpp") == 0 ){
			VESA_BPP = 16;
		}else if( strcmp(argv[index], "-24bpp") == 0 ){
			VESA_BPP = 24;
		}else if( strcmp(argv[index], "-32bpp") == 0 ){
			VESA_BPP = 32;
		}
		index += 1;
	}

	// scrmode.log
	if((g_scrmodelog = fopen("scrmode.log", "a")) == NULL){
			printf("cannot open file scrmode.log\n");
			return -1;
	}

	if(DetectSVGA15KHz() == FALSE){
		printf("Device not found.\n");
		return -1;
	}

	GetCRTCPARAM(&cp, firstscrmode);
	AdjustCRTCPARAM(&cp);

	g_undo_cp1 = cp;
	g_undo_cp2 = cp;

	allegro_init();
	install_keyboard();

	if( InitGfxMode() == FALSE ){
		goto error;
		return -1;
	}

	cursorpos = 0;
	SetMode15K(&cp);
	WriteGrid(VESA_WIDTH, VESA_HEIGHT, grid_intensity, draw_edge_line,
			  cp.HDisp - 1, cp.VDisp - 1, cp.interlace, cp.doublescan);
	DispText(&cp, cursorpos);

	// main loop
	while(1){
		keydata = readkey() >> 8;
		if(keydata == KEY_ESC) break;
		if(keydata == KEY_UP)	cursorpos -= 1;
		if(keydata == KEY_DOWN) cursorpos += 1;
		if(cursorpos < 0) cursorpos  = MAX_PARAM - 1;
		if(cursorpos > (MAX_PARAM - 1)) cursorpos = 0;

		if(keydata == KEY_LEFT)
			IncParam(-1, &cp, cursorpos);
		if(keydata == KEY_RIGHT)
			IncParam( 1, &cp, cursorpos);
		if(keydata == KEY_PGDN)
			IncParam(-10, &cp, cursorpos);
		if(keydata == KEY_PGUP)
			IncParam( 10, &cp, cursorpos);

		if(keydata == KEY_INSERT){
			grid_intensity = 1;
			WriteGrid(VESA_WIDTH, VESA_HEIGHT, grid_intensity, draw_edge_line,
					  cp.HDisp - 1, cp.VDisp - 1, cp.interlace, cp.doublescan);
		}
		if(keydata == KEY_DEL){
			grid_intensity = 0;
			WriteGrid(VESA_WIDTH, VESA_HEIGHT, grid_intensity, draw_edge_line,
					  cp.HDisp - 1, cp.VDisp - 1, cp.interlace, cp.doublescan);
		}
		if(keydata == KEY_HOME){
			draw_edge_line = 1;
			WriteGrid(VESA_WIDTH, VESA_HEIGHT, grid_intensity, draw_edge_line,
					  cp.HDisp - 1, cp.VDisp - 1, cp.interlace, cp.doublescan);
		}
		if(keydata == KEY_END){
			draw_edge_line = 0;
			WriteGrid(VESA_WIDTH, VESA_HEIGHT, grid_intensity, draw_edge_line,
					  cp.HDisp - 1, cp.VDisp - 1, cp.interlace, cp.doublescan);
		}


		if(keydata == KEY_ENTER){
			AdjustCRTCPARAM(&cp);
			SetMode15K(&cp);
			WriteGrid(VESA_WIDTH, VESA_HEIGHT, grid_intensity, draw_edge_line,
					  cp.HDisp - 1, cp.VDisp - 1, cp.interlace, cp.doublescan);
		}
		if(keydata == KEY_U){
			cp = g_undo_cp2;
			AdjustCRTCPARAM(&cp);
			SetMode15K(&cp);
			WriteGrid(VESA_WIDTH, VESA_HEIGHT, grid_intensity, draw_edge_line,
					  cp.HDisp - 1, cp.VDisp - 1, cp.interlace, cp.doublescan);
		}
		if(keydata == KEY_BACKSPACE){
			cp = g_undo_cp1;
			AdjustCRTCPARAM(&cp);
		}

		if(keydata == KEY_F1  || keydata == KEY_F2	||
		   keydata == KEY_F3  || keydata == KEY_F4	||
		   keydata == KEY_F5  || keydata == KEY_F6	||
		   keydata == KEY_F7  || keydata == KEY_F8	||
		   keydata == KEY_F9  || keydata == KEY_F10 ||
		   keydata == KEY_F11 || keydata == KEY_F12)
		{
			int mode;

			switch(keydata){
				case KEY_F1:  mode = GET_320x240_60_15KHz; break;
				case KEY_F2:  mode = GET_512x384_60_24KHz; break;
				case KEY_F3:  mode = GET_640x480_60_31KHz; break;
				case KEY_F4:  mode = GET_512x384_60_31KHz; break;
				case KEY_F5:  mode = GET_F5;  break;
				case KEY_F6:  mode = GET_F6;  break;
				case KEY_F7:  mode = GET_F7;  break;
				case KEY_F8:  mode = GET_F8;  break;
				case KEY_F9:  mode = GET_F9;  break;
				case KEY_F10: mode = GET_F10; break;
				case KEY_F11: mode = GET_F11; break;
				case KEY_F12: mode = GET_F12; break;
				default:	  mode = GET_640x480_60_31KHz;
			}
			GetCRTCPARAM(&cp, mode);
			AdjustCRTCPARAM(&cp);
			SetMode15K(&cp);
			WriteGrid(VESA_WIDTH, VESA_HEIGHT, grid_intensity, draw_edge_line,
					  cp.HDisp - 1, cp.VDisp - 1, cp.interlace, cp.doublescan);
		}

		if(keydata == KEY_SPACE){
			AdjustCRTCPARAM(&cp);
			DispCP(&cp);
		}

		AdjustCRTCPARAM(&cp);
		DispText(&cp, cursorpos);


	}


error:
	fclose(g_scrmodelog);
	ExitSVGA15KHz();
	ExitGfxMode();
	return 0;
}

END_OF_MAIN();
