#include "mamalleg.h"
#include "driver.h"
#include <math.h>
#include <pc.h>
#include "gen15khz.h"
#include "mode15k.h"
#include "smntrlst.h"



char svga15kdrv_monitorname[256];


static void vesa_scroll(int x, int y);

static int usepcmonitor = 0;

static CRTCPARAM rcp;
static char orientation_string[4][15] = {
		{ "horz" }, { "vert" },
		{ "horz_ystretch" }, { "vert_ystretch" }
};


/* from blit.c */
extern int use_y_stretch;

/* from video.c */
extern SVGA15KHZDRIVER *SVGA15KHzdriver;


extern int svga_driver_monitor_type;

extern int video_depth, video_attributes, video_orientation;
extern float video_fps;
extern int width_for_mode15k, height_for_mode15k;

extern int scanlines;
extern int stretch;
extern int center_x;
extern int center_y;
extern int video_sync;
extern int use_triplebuf;
extern int gfx_xoffset;
extern int gfx_yoffset;

extern int gfx_width;
extern int gfx_height;
extern int _gfx_width;
extern int _gfx_height;
extern int interlaceratio;
extern float fps_value;

extern int horizontal_sync_delay;
extern int horizontal_disp_delay;

extern int pcmonitorlow_h;
extern int pcmonitorhigh_h;
extern int pcmonitorlow_v;
extern int pcmonitorhigh_v;
extern int usespecial15KHz;
extern int use_strict_mode_only;
extern int useinterlace;

extern int use_svga15khz_panning;

extern int use_crtc_data_base;
extern int crtc_data_base_write_only;

extern SCRLIST ScrModeList[];
extern SCRLIST AutoResolutionList[];
extern SCRLIST YStretchList[];
extern SCRLIST YStretchAutoList[];

static void AdjustCenter(CRTCPARAM *cp);



static int set_crtc_param_cfg(char *gamename, char *monitorname,
								char *keyname, CRTCPARAM *cp)
{
	char sbuff_regs[512];
	char sbuff_misc[512];
	char temp[2048];

	if(strlen(keyname) >= 128) return 0;
	if(strlen(monitorname) >= 128) return 0;

	strcpy(sbuff_regs, monitorname);
	strcpy(sbuff_misc, monitorname);
	strcat(sbuff_regs, "_");
	strcat(sbuff_misc, "_");
	strcat(sbuff_regs, keyname);
	strcat(sbuff_misc, keyname);
	strcat(sbuff_regs, "_regs");
	strcat(sbuff_misc, "_misc");

	push_config_state();
	set_config_file("crtc_db.cfg");
	sprintf(temp, "%d %d %d %d %d %d %d %d %d %d %d %d",
			cp->HTotal	,
			cp->HDisp	,
			cp->HBStart ,
			cp->HSStart ,
			cp->HSEnd	,
			cp->HBEnd	,
			cp->VTotal	,
			cp->VDisp	,
			cp->VBStart ,
			cp->VSStart ,
			cp->VSEnd	,
			cp->VBEnd
	);
	set_config_string(gamename, sbuff_regs, temp);

	sprintf(temp, "%d %d %d %d %d %d %d %ld %ld %ld",
			cp->width		   ,
			cp->height		   ,
			cp->hpolarity	   ,
			cp->vpolarity	   ,
			cp->doublescan	   ,
			cp->interlace	   ,
			cp->interlaceadjust,
			cp->dotclockHz	   ,
			cp->horzHz		   ,
			cp->vert001Hz
	);
	set_config_string(gamename, sbuff_misc, temp);

	pop_config_state();
	return 1;
}


static int get_crtc_param_cfg(char *gamename, char *monitorname,
								char *keyname, CRTCPARAM *cp)
{
	char sbuff_regs[512];
	char sbuff_misc[512];
	int d0;
	int d1;
	char **a0;
	char **a1;

	if(strlen(keyname) >= 128) return 0;
	if(strlen(monitorname) >= 128) return 0;

	strcpy(sbuff_regs, monitorname);
	strcpy(sbuff_misc, monitorname);
	strcat(sbuff_regs, "_");
	strcat(sbuff_misc, "_");
	strcat(sbuff_regs, keyname);
	strcat(sbuff_misc, keyname);
	strcat(sbuff_regs, "_regs");
	strcat(sbuff_misc, "_misc");

	push_config_state();
	set_config_file("crtc_db.cfg");

	a0 = get_config_argv(gamename, sbuff_regs, &d0);
	a1 = get_config_argv(gamename, sbuff_misc, &d1);

	if(d0 == 0 || d1 == 0) return 0;	// not found
	if(d0 != 12) return 0;				// invalid
	if(d1 != 10) return 0;				// invalid

	a0 = get_config_argv(gamename, sbuff_regs, &d0);
	cp->HTotal	= atoi( *(a0 +	0));
	cp->HDisp	= atoi( *(a0 +	1));
	cp->HBStart = atoi( *(a0 +	2));
	cp->HSStart = atoi( *(a0 +	3));
	cp->HSEnd	= atoi( *(a0 +	4));
	cp->HBEnd	= atoi( *(a0 +	5));
	cp->VTotal	= atoi( *(a0 +	6));
	cp->VDisp	= atoi( *(a0 +	7));
	cp->VBStart = atoi( *(a0 +	8));
	cp->VSStart = atoi( *(a0 +	9));
	cp->VSEnd	= atoi( *(a0 + 10));
	cp->VBEnd	= atoi( *(a0 + 11));

	a1 = get_config_argv(gamename, sbuff_misc, &d1);
	cp->width			= atoi( *(a1 + 0));
	cp->height			= atoi( *(a1 + 1));
	cp->hpolarity		= atoi( *(a1 + 2));
	cp->vpolarity		= atoi( *(a1 + 3));
	cp->doublescan		= atoi( *(a1 + 4));
	cp->interlace		= atoi( *(a1 + 5));
	cp->interlaceadjust = atoi( *(a1 + 6));
	cp->dotclockHz = atol( *(a1 + 7));
	cp->horzHz	   = atol( *(a1 + 8));
	cp->vert001Hz  = atol( *(a1 + 9));

	pop_config_state();

	if(cp->HTotal	  == 0) return 0;
	if(cp->HDisp	  == 0) return 0;
	if(cp->HBStart	  == 0) return 0;
	if(cp->HSStart	  == 0) return 0;
	if(cp->HSEnd	  == 0) return 0;
	if(cp->HBEnd	  == 0) return 0;
	if(cp->VTotal	  == 0) return 0;
	if(cp->VDisp	  == 0) return 0;
	if(cp->VBStart	  == 0) return 0;
	if(cp->VSStart	  == 0) return 0;
	if(cp->VSEnd	  == 0) return 0;
	if(cp->VBEnd	  == 0) return 0;
	if(cp->dotclockHz == 0) return 0;
	if(cp->horzHz	  == 0) return 0;
	if(cp->vert001Hz  == 0) return 0;

	return 1;
}






void LogAttributes(void)
{
	int d0;

	logerror("video attributes:\n");

	if(video_attributes & 0x01){
		logerror("\tVIDEO_TYPE_VECTOR\n");
	}else{
		logerror("\tVIDEO_TYPE_RASTER\n");
	}
	if(video_attributes & VIDEO_DUAL_MONITOR)
		logerror("\tDUAL_MONITOR\n");

	logerror("\tASPECT_RATIO ");
	d0 = video_attributes & VIDEO_PIXEL_ASPECT_RATIO_MASK;
	switch(d0){
		case VIDEO_PIXEL_ASPECT_RATIO_1_1:
			logerror("1:1\n");
			break;
		case VIDEO_PIXEL_ASPECT_RATIO_1_2:
			logerror("1:2\n");
			break;
		case VIDEO_PIXEL_ASPECT_RATIO_2_1:
			logerror("2:1\n");
			break;
	}
}



void LogScrGroup(int group)
{
	DWORD d0;
	int d1;

	d0 = 0x00000001;
	d1 = 32;
	while(d1 > 0){
		switch(group & d0){
			case STRICT15KHz:
				logerror("STRICT15KHz ");
				break;
			case STRICTTV15KHz:
				logerror("STRICTTV15KHz ");
				break;
			case TV15KHz:
				logerror("TV15KHz ");
				break;
			case MODE15KHz:
				logerror("MODE15KHz ");
				break;
			case STRICT24KHz:
				logerror("STRICT24KHz ");
				break;
			case MODE24KHz:
				logerror("MODE24KHz ");
				break;
			case MODE31KHz:
				logerror("MODE31KHz ");
				break;
			case PAL15KHz:
				logerror("PAL15KHz ");
				break;
			case SPECIAL15KHz:
				logerror("SPECIAL15KHz ");
				break;
			case PCMONITOR:
				logerror("PCMONITOR ");
				break;
		}
		d0 = d0 << 1;
		d1 -= 1;
	}
}

void LogScrGroupFor(int group)
{
	logerror("for ");
	LogScrGroup(group);
}

void LogScrOrientation(int orientation)
{
	logerror("for");

	if(orientation & SCREEN_HORZ){
		logerror(" HORZ");
	}
	if(orientation & SCREEN_VERT){
		logerror(" VERT");
	}
	if(orientation & DUAL_MONITOR){
		logerror(" DUAL_MONITOR");
	}
	logerror(", ");
}


static void log_scr(SCRLIST *ListPtr)
{
	logerror("found %dx%d (%dx%d) H:%.2fKHz V:%.2fHz, ",
					ListPtr->cp->width,
					ListPtr->cp->height,
					ListPtr->forx,
					ListPtr->fory,
					(float)ListPtr->cp->horzHz / 1000,
					(float)ListPtr->cp->vert001Hz / 100);
	LogScrOrientation(ListPtr->orientation);
	LogScrGroupFor(ListPtr->group);
	logerror("\n");
}

static void log_scr_found_but(SCRLIST *ListPtr, long frame_rate)
{
	logerror("found %dx%d (%dx%d) H:%.2fKHz V:%.2fHz, ",
					ListPtr->cp->width,
					ListPtr->cp->height,
					ListPtr->forx,
					ListPtr->fory,
					(float)ListPtr->cp->horzHz / 1000,
					(float)ListPtr->cp->vert001Hz / 100);
	LogScrOrientation(ListPtr->orientation);
	LogScrGroupFor(ListPtr->group);
	logerror("\n");
	logerror("but not V:%.2fHz\n", (float)frame_rate / 100);
}

static void log_scr_found_but_interlaced_mode(SCRLIST *ListPtr)
{
	logerror("found %dx%d (%dx%d) H:%.2fKHz V:%.2fHz, ",
					ListPtr->cp->width,
					ListPtr->cp->height,
					ListPtr->forx,
					ListPtr->fory,
					(float)ListPtr->cp->horzHz / 1000,
					(float)ListPtr->cp->vert001Hz / 100);
	LogScrOrientation(ListPtr->orientation);
	LogScrGroupFor(ListPtr->group);
	logerror("\n");
	logerror("but interlaced mode\n");
}




static int search_scr(SCRLIST *ListPtr, SCRLIST **scrptr,
						int display_columns, int display_lines, long frame_rate,
						int vsyncflag, DWORD scrmode, DWORD orientation,
						int autoresflag)
{
//
// if found, return != 0 and scrptr is valid
//
	int idx;
	int found;
	int flag;

	if(vsyncflag == 0){
		logerror("refresh rate comparison off\n");
	}else{
		logerror("refresh rate comparison on\n");
	}

	found = 0;
	idx = 0;
	while(1){
		if( (ListPtr + idx)->forx == 0 || (ListPtr + idx)->fory == 0) {
			logerror("%dx%d @ %.2fHz ", display_columns, display_lines,
										(float)frame_rate/100);
			LogScrOrientation(orientation);
			LogScrGroup(scrmode);
			logerror("mode not found\n");
			break;
		}
		if(autoresflag == 0){
			flag =	(((ListPtr + idx)->group & scrmode) != 0)	&&
					(((ListPtr + idx)->orientation & orientation) == orientation);
			if((ListPtr + idx)->forx != -1){
				flag = flag && ((ListPtr + idx)->forx == display_columns);
			}
			if((ListPtr + idx)->fory != -1){
				flag = flag && ((ListPtr + idx)->fory == display_lines);
			}
		}else{
			flag =	(((ListPtr + idx)->group & scrmode) != 0)	&&
					(((ListPtr + idx)->orientation & orientation) == orientation);
			if((ListPtr + idx)->forx != -1){
				flag = flag && ((ListPtr + idx)->forx >= display_columns);
			}
			if((ListPtr + idx)->fory != -1){
				flag = flag && ((ListPtr + idx)->fory >= display_lines);
			}
		}
		if(flag)
		{
			if((useinterlace == 0) && ((ListPtr + idx)->cp->interlace != 0)){
				log_scr_found_but_interlaced_mode(ListPtr + idx);
			}else{
				if(vsyncflag == 0){
					log_scr(ListPtr + idx);
					found = 1;
					break;
				}else{
					long ld0;
					long temp;

					ld0 = (ListPtr + idx)->cp->vert001Hz;
					temp = frame_rate;
					while(1){
						if( (temp * 1.9) < ld0 ){
							temp *= 2;
						}else{
							break;
						}
					}
					ld0 = abs(ld0 - temp);
					if(ld0 < 10){
						log_scr(ListPtr + idx);
						found = 1;
						break;
					}else{
						log_scr_found_but(ListPtr + idx, frame_rate);
					}
				}
			}
		}
		idx += 1;
	}

	if(found != 0){
		*scrptr = ListPtr + idx;
		return 1;
	}
	return 0;
}




static DWORD get_scr_atrb(int monitor_type)
{
	DWORD scrmode;

	switch(monitor_type){
		case SVGA_DRIVER_MONITOR_TYPE_STANDARD_PC:
			scrmode = PCMONITOR;
			logerror("mode15k: STANDARD PC MONITOR\n");
			usepcmonitor = 1;
			break;

		case SVGA_DRIVER_MONITOR_TYPE_NTSC:
			scrmode = STRICT15KHz;
			logerror("mode15k: NTSC MONITOR\n");
			logerror("mode15k: strict 15.75KHz only\n");
			break;

		case SVGA_DRIVER_MONITOR_TYPE_TV15:
			logerror("mode15k: TV 15KHz 60Hz MONITOR\n");
			logerror("mode15k: 224 lines modes have priority\n");
			if(use_strict_mode_only != 0){
				scrmode = STRICTTV15KHz;
			}else{
				scrmode = STRICTTV15KHz | TV15KHz;
			}
			break;

		case SVGA_DRIVER_MONITOR_TYPE_ARCADE15:
			logerror("mode15k: 15KHz ARCADE MONITOR\n");
			if(use_strict_mode_only != 0){
				scrmode = STRICT15KHz | STRICTTV15KHz;
			}else{
				scrmode = STRICT15KHz | STRICTTV15KHz | TV15KHz | MODE15KHz;
				if(usespecial15KHz != 0){
					scrmode |= SPECIAL15KHz;
				}
			}
			break;

		case SVGA_DRIVER_MONITOR_TYPE_ARCADE24:
			scrmode = MODE24KHz | STRICT24KHz;
			logerror("mode15k: 24KHz MONITOR\n");
			break;

		case SVGA_DRIVER_MONITOR_TYPE_ARCADE31:
			scrmode = MODE31KHz;
			logerror("mode15k: 31KHz MONITOR\n");
			break;

		case SVGA_DRIVER_MONITOR_TYPE_ARCADE1524:
			logerror("mode15k: 15KHz 24KHz ARCADE MONITOR\n");
			if(use_strict_mode_only != 0){
				scrmode  = STRICT15KHz | STRICTTV15KHz | STRICT24KHz | MODE24KHz;
			}else{
				scrmode  = MODE15KHz | STRICT15KHz | STRICTTV15KHz | TV15KHz |
						   MODE24KHz | STRICT24KHz;
				if(usespecial15KHz != 0){
					scrmode |= SPECIAL15KHz;
				}
			}
			break;

		case SVGA_DRIVER_MONITOR_TYPE_ARCADE152431:
			logerror("mode15k: 15KHz 24KHz 31KHz ARCADE MONITOR\n");
			if(use_strict_mode_only != 0){
				scrmode  = STRICT15KHz | STRICTTV15KHz;
				scrmode |= MODE24KHz | STRICT24KHz;
				scrmode |= MODE31KHz;
			}else{
				scrmode  = MODE15KHz | STRICT15KHz | STRICTTV15KHz | TV15KHz;
				scrmode |= MODE24KHz | STRICT24KHz;
				scrmode |= MODE31KHz;
				if(usespecial15KHz != 0){
					scrmode |= SPECIAL15KHz;
				}
			}
			break;

		case SVGA_DRIVER_MONITOR_TYPE_ARCADE2431:
			if(use_strict_mode_only != 0){
				scrmode  = MODE24KHz | STRICT24KHz;
				scrmode |= MODE31KHz;
			}else{
				scrmode  = MODE24KHz | STRICT24KHz;
				scrmode |= MODE31KHz;
			}
			logerror("mode15k: 24KHz 31KHz ARCADE MONITOR\n");
			break;

		default:
			scrmode  = 0;
			logerror("mode15k: wrong monitor\n");
	}

	logerror("mode15k: scrmode ");
	LogScrGroup(scrmode);
	logerror("\n");

	return scrmode;
}






int GetMode15KHz(CRTCPARAM *cp)
{
	SCRLIST *ScrPtr;
	DWORD scrmode;
	int found;
	int display_columns;
	int display_lines;
	DWORD orientation;
	int vsyncflag;
	int vesares_x;
	int vesares_y;
	long frame_rate;
	int d0;

	found = 0;
	if(use_crtc_data_base != 0 && crtc_data_base_write_only == 0){
		if((video_orientation & ORIENTATION_SWAP_XY) == 0)
			d0 = 0; 	// horz
		else
			d0 = 1; 	// vert
		if(use_y_stretch != 0) d0 += 2;
		found = get_crtc_param_cfg((char*)Machine->gamedrv->name, svga15kdrv_monitorname,
									&orientation_string[d0][0], cp);
		logerror("\ncrtc_db.cfg : [%s] for %s ", (char*)Machine->gamedrv->name,
											&orientation_string[d0][0]);
		if(found != 0) logerror("found\n\n");
		else		   logerror("not found\n\n");
	}

	if(found == 0){
		vesares_x = width_for_mode15k;
		vesares_y = height_for_mode15k;

		if(fps_value == 0.0){
			frame_rate = (long)((video_fps * 100) + 0.005);
			logerror("frame_rate = video_fps (= %.2f)\n",
										(float)frame_rate/100);
		}else{
			frame_rate = (long)((fps_value * 100) + 0.005);
			logerror("frame_rate = fps_value (= %.2f)\n",
										(float)frame_rate/100);
		}
		orientation = 0;
		LogAttributes();
		if((video_attributes & VIDEO_DUAL_MONITOR) != 0){
			orientation |= DUAL_MONITOR;
		}
		if((video_orientation & ORIENTATION_SWAP_XY) == 0){
			orientation |= SCREEN_HORZ;
			logerror("orientation : Horz\n");
		}else{
			orientation |= SCREEN_VERT;
			logerror("orientation : Vert\n");
		}

		logerror("\n");
		scrmode = get_scr_atrb(svga_driver_monitor_type);
		if(scrmode == 0) return 0;
		logerror("\n");

		logerror("_gfx_width.... %d\n", _gfx_width);
		logerror("_gfx_height... %d\n", _gfx_height);

		if(_gfx_width == 0 && _gfx_height == 0){
			display_columns = vesares_x;
			display_lines	= vesares_y;
		}else{
			display_columns = _gfx_width;
			display_lines	= _gfx_height;
		}
		logerror("request  %dx%d ", display_columns, display_lines);
		logerror("%.2fHz\n", (float)frame_rate/100);

		while(1){
			SCRLIST *normallist;
			SCRLIST *autolist;

			if(use_y_stretch == 0){
				normallist = ScrModeList;
				autolist   = AutoResolutionList;
			}else{
				logerror("y-stretch is on\n");
				normallist = YStretchList;
				autolist   = YStretchAutoList;
			}
			logerror("###NORMAL LIST###\n");
			vsyncflag = 1;
			found = search_scr(normallist, &ScrPtr,
								display_columns, display_lines, frame_rate,
								vsyncflag, scrmode, orientation, 0);
			if(found != 0) break;
			if(vsyncflag != 0){
				found = search_scr(normallist, &ScrPtr,
									 display_columns, display_lines, frame_rate,
									 0, scrmode, orientation, 0);
				if(found != 0) break;
			}
			logerror("###AUTORES LIST###\n");
			found = search_scr(autolist, &ScrPtr,
								display_columns, display_lines, frame_rate,
								vsyncflag, scrmode, orientation, 1);
			if(found != 0) break;
			if(vsyncflag != 0){
				found = search_scr(autolist, &ScrPtr,
									 display_columns, display_lines, frame_rate,
									 0, scrmode, orientation, 1);
				if(found != 0) break;
			}
			break;
		}

		if(found == 0){
			logerror("not found\n");
			return 0;
		}

		*cp = *(ScrPtr->cp);
		AdjustCenter(cp);

		while(usepcmonitor != 0){
			long vert001Hz;
			long horzHz;
			long dotclockHz;
			int limitlow_v;
			int limithigh_v;
			int limitlow_h;
			int limithigh_h;

			if((pcmonitorlow_v == 0) || (pcmonitorhigh_v == 0) ||
			   (pcmonitorlow_h == 0) || (pcmonitorhigh_h == 0))
			{
				limitlow_v	= 60;					// Hz
				limithigh_v = 75;					// Hz
				limitlow_h	= 31;					// KHz
				limithigh_h = 50;					// KHz
			}else{
				limitlow_v	= pcmonitorlow_v;		// Hz
				limithigh_v = pcmonitorhigh_v;		// Hz
				limitlow_h	= pcmonitorlow_h;		// KHz
				limithigh_h = pcmonitorhigh_h;		// KHz
			}

			logerror("PC MONITOR MODE\n");
			vert001Hz  = frame_rate;
			logerror("V %dHz\n", vert001Hz/100);
			logerror("pc monitor low  limit V %dHz\n", limitlow_v);
			logerror("pc monitor high limit V %dHz\n", limithigh_v);
			while((vert001Hz/100) <= limithigh_v){
				if( (limitlow_v <= (vert001Hz/100)) &&
					((vert001Hz/100) <= limithigh_v) )
				{
					break;
				}
				vert001Hz *= 2;
				logerror("x2 V rate\n");
			}
			if((vert001Hz/100) > limithigh_v){
				logerror("V %dHz > %dHz\n", vert001Hz/100, limithigh_v);
				logerror("Using default 60Hz mode\n");
				break;
			}
			dotclockHz = CalcDotclock(vert001Hz, cp->HTotal, cp->VTotal,
													cp->interlace);
			horzHz	   = CalcHorizontalHz(dotclockHz, cp->HTotal);
			logerror("H %dKHz\n", horzHz/1000);
			logerror("pc monitor low  limit H %dKHz\n", limitlow_h);
			logerror("pc monitor high limit H %dKHz\n", limithigh_h);
			if( !((limitlow_h <= (horzHz/1000)) && ((horzHz/1000) <= limithigh_h)) ){
				logerror("Using default 60Hz mode\n");
				break;
			}
			cp->vert001Hz  = vert001Hz;
			cp->horzHz	   = horzHz;
			cp->dotclockHz = dotclockHz;
			logerror("Set H:%.2fKHz V:%.2fHz dotclock %.4f MHz\n",
							(float)cp->horzHz / 1000,
							(float)cp->vert001Hz / 100,
							(float)(cp->dotclockHz / 1000) / 1000.0);
			break;
		}
	}

	gfx_width  = cp->width;
	gfx_height = cp->height;
	use_svga15khz_panning = 1;
	setinterlaceflag(0);

	if(use_crtc_data_base != 0){
		/* save crtcparam to crtc data base */
		if((video_orientation & ORIENTATION_SWAP_XY) == 0)
			d0 = 0; 	// horz
		else
			d0 = 1; 	// vert
		if(use_y_stretch != 0) d0 += 2;
		set_crtc_param_cfg((char*)Machine->gamedrv->name, svga15kdrv_monitorname,
								&orientation_string[d0][0], cp);
	}

	return 1;
}




static void vesa_scroll(int x, int y)
{
	__dpmi_regs _dpmi_reg;

	_dpmi_reg.x.ax = 0x4F07;
	_dpmi_reg.x.bx = 0;
	_dpmi_reg.x.cx = x;
	_dpmi_reg.x.dx = y;
	__dpmi_int(0x10, &_dpmi_reg);
}

static void AdjustCenter(CRTCPARAM *cp)
{
	int d0;

	if(center_x < 0){
		if((cp->HSStart + center_x) < (cp->HDisp + 8)){
			d0 = cp->HSStart;
			cp->HSStart = cp->HDisp + 8;
			cp->HSEnd  -= (d0 - cp->HSStart);
		}else{
			cp->HSStart = cp->HSStart + center_x;
			cp->HSEnd	= cp->HSEnd   + center_x;
		}
	}else{
		if((cp->HSEnd + center_x) > (cp->HTotal - 16)){
			d0 = cp->HSEnd;
			cp->HSEnd	 = cp->HTotal - 16;
			cp->HSStart += (cp->HSEnd - d0);
		}else{
			cp->HSEnd	= cp->HSEnd   + center_x;
			cp->HSStart = cp->HSStart + center_x;
		}
	}

	if(center_y < 0){
		if((cp->VSStart + center_y) < (cp->VDisp + 1)){
			d0 = cp->VSStart;
			cp->VSStart = cp->VDisp + 1;
			cp->VSEnd  -= (d0 - cp->VSStart);
		}else{
			cp->VSStart = cp->VSStart + center_y;
			cp->VSEnd	= cp->VSEnd   + center_y;
		}
	}else{
		if((cp->VSEnd + center_y) > (cp->VTotal - 2)){
			d0 = cp->VSEnd;
			cp->VSEnd	 = cp->VTotal - 2;
			cp->VSStart += (cp->VSEnd - d0);
		}else{
			cp->VSEnd	= cp->VSEnd   + center_y;
			cp->VSStart = cp->VSStart + center_y;
		}
	}
}


void Centering(CRTCPARAM *cp, int dx, int dy)
{
	if(dx < 0){
		if((cp->HSStart + dx) > cp->HDisp){
			cp->HSStart += dx;
			cp->HSEnd	+= dx;
		}
	}else{
		if((cp->HSEnd + dx) < (cp->HTotal - 8)){
			cp->HSStart += dx;
			cp->HSEnd	+= dx;
		}
	}

	if(dy < 0){
		if((cp->VSStart + dy) > cp->VDisp){
			cp->VSStart += dy;
			cp->VSEnd	+= dy;
		}
	}else{
		if((cp->VSEnd + dy) < (cp->VTotal - 1)){
			cp->VSStart += dy;
			cp->VSEnd	+= dy;
		}
	}
}




#if 0
void GetCrtcParam(CRTCPARAM *cp, MiscParam *mc, _Register *rg, int registerlen)
{
	BYTE port[0x20];
	int d0;
	int d1;

	d0 = 0;
	while(d0 < registerlen){
		port[ rg[d0].index ] = rg[d0].value;
		d0 += 1;
	}


/*										*/
/*				 CRTC					*/
/*										*/

	cp->HTotal	= (port[0x00] + 5) * 8;
	cp->HDisp	= (port[0x01] + 1) * 8;

	d0 = port[0x02];
	cp->HBStart = d0 * 8;
	d1 = (port[0x03] & 0x1F) | BitMov(0, 5, port[0x05], 7);
	while( (d0 & 0x3F) != (d1 & 0x3F) ){
		d0 += 1;
	}
	cp->HBEnd	= d0 * 8;

	d0 = port[0x04];
//	  cp->HSStart = d0 * 8;
	cp->HSStart = d0 * 8 + mc->hskew;
	d1 = (port[0x05] & 0x1F);
	while( (d0 & 0x1F) != (d1 & 0x1F) ){
		d0 += 1;
	}
	cp->HSEnd	= d0 * 8;

	cp->VTotal	= (port[0x06] | BitMov(0, 8, port[0x07], 0) |
								BitMov(0, 9, port[0x07], 5) ) + 2;
	cp->VDisp	= (port[0x12] | BitMov(0, 8, port[0x07], 1) |
								BitMov(0, 9, port[0x07], 6) ) + 1;
	cp->VBStart = (port[0x15] | BitMov(0, 8, port[0x07], 3) |
								BitMov(0, 9, port[0x09], 5) );
	d0 = cp->VBStart;
	d1 = port[0x16];
	while( (d0 & 0xFF) != (d1 & 0xFF) ){
		d0 += 1;
	}

//.    cp->VBEnd = d0;
//.    cp->VBEnd = cp->VTotal - 16;


	cp->VSStart = (port[0x10] | BitMov(0, 8, port[0x07], 2) |
								BitMov(0, 9, port[0x07], 7) );
	d0 = cp->VSStart;
	d1 = port[0x11] & 0x0F;
	while( (d0 & 0x0F) != (d1 & 0x0F) ){
		d0 += 1;
	}
	cp->VSEnd = d0;
//.
	cp->VBEnd = cp->VSEnd + 1;
//.



/*										*/
/*				 MISC					*/
/*										*/

	cp->width  = mc->x;
	cp->height = mc->y;
	cp->depth  = 0;

	cp->dotclockHz = (long)(mc->dotclock * 1000 * 1000);
	cp->horzHz	   = (long)(mc->horzKHz * 1000);
	cp->vert001Hz  = (long)((mc->vertHz + 0.005) * 100);

	cp->hpolarity  = mc->hpolarity;
	cp->vpolarity  = mc->vpolarity;
	cp->doublescan = (port[0x09] >> 7) & 0x01;
	cp->interlace  = mc->interlace;

	cp->interlaceadjust = 0;
}
#endif




void SetGenericVGARegister(CRTCPARAM *cp)
{
	int d0;
	int d1;
	int HDispSkew;
	int HSyncDelay;

	HSyncDelay = horizontal_sync_delay & 0x03;
	HDispSkew  = horizontal_disp_delay & 0x03;

	// Unlock
	SetCrtcReg(0x11, BitMov(GetCrtcReg(0x11), 7, 0, 0));

	// HTotal
	d0 = cp->HTotal;
	d0 = (d0 / 8) - 5;
	SetCrtcReg(0x00, d0 & 0xFF);

	// HDisp
	d0 = cp->HDisp;
	d0 = (d0 / 8) - 1;
	SetCrtcReg(0x01, d0 & 0xFF);

	// HBlankStart
	d0 = cp->HBStart;
	d0 = d0 / 8;
	SetCrtcReg(0x02, d0 & 0xFF);

	// HBlankEnd
	d0 = cp->HBEnd;
	d0 = d0 / 8;
	d1 = GetCrtcReg(0x03);
	d1 = BitMov(d1, 0, d0, 0);
	d1 = BitMov(d1, 1, d0, 1);
	d1 = BitMov(d1, 2, d0, 2);
	d1 = BitMov(d1, 3, d0, 3);
	d1 = BitMov(d1, 4, d0, 4);
	d1 = BitMov(d1, 5, HDispSkew, 0);	// horizontal disp skew
	d1 = BitMov(d1, 6, HDispSkew, 1);	// horizontal disp skew
	d1 = BitMov(d1, 7, 1, 0);			// compatible read
	SetCrtcReg(0x03, d1);
	SetCrtcReg(0x05, BitMov(GetCrtcReg(0x05), 7, d0, 5));

	// HSyncStart
	d0 = cp->HSStart;
	d0 = d0 / 8;
	SetCrtcReg(0x04, d0 & 0xFF);

	// HSyncEnd
	d0 = cp->HSEnd;
	d0 = d0 / 8;
	d1 = GetCrtcReg(0x05);
	d1 = BitMov(d1, 0, d0, 0);
	d1 = BitMov(d1, 1, d0, 1);
	d1 = BitMov(d1, 2, d0, 2);
	d1 = BitMov(d1, 3, d0, 3);
	d1 = BitMov(d1, 4, d0, 4);
	d1 = BitMov(d1, 5, HSyncDelay, 0);
	d1 = BitMov(d1, 6, HSyncDelay, 1);
	SetCrtcReg(0x05, d1);

	// VTotal
	d0 = cp->VTotal;
	d0 = d0 - 2;
	SetCrtcReg(0x06, d0 & 0xFF);
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 0, d0, 8));
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 5, d0, 9));

	// VDisp
	d0 = cp->VDisp;
	d0 = d0 - 1;
	SetCrtcReg(0x12, d0 & 0xFF);
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 1, d0, 8));
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 6, d0, 9));


	// VBlankStart
	d0 = cp->VBStart;
	SetCrtcReg(0x15, d0 & 0xFF);
	SetCrtcReg(0x07, BitMov(GetCrtcReg(0x07), 3, d0, 8));
	SetCrtcReg(0x09, BitMov(GetCrtcReg(0x09), 5, d0, 9));

	// VBlankEnd
	d0 = cp->VBEnd;
	SetCrtcReg(0x16, d0 & 0xFF);

	// VSyncStart
	d0 = cp->VSStart;
	SetCrtcReg(0x10, d0 & 0xFF);
	d1 = GetCrtcReg(0x07);
	d1 = BitMov(d1, 2, d0,	8);  // bit 8
	d1 = BitMov(d1, 7, d0,	9);  // bit 9
	SetCrtcReg(0x07, d1);

	// VSyncEnd
	d0 = cp->VSEnd;
	d0 = d0 & 0x0F;
	d1 = GetCrtcReg(0x11) & 0xF0;
	d1 = d1 | d0;
	SetCrtcReg(0x11, d1);

	// Double Scan Dis.
	SetCrtcReg(0x09, BitMov(GetCrtcReg(0x09), 7, 0, 0));

}



long CalcDotclock(long Vert001Hz, int HTotal, int VTotal, int interlace)
{
	double f0;

	f0 = (double)VTotal;
	if(interlace != 0){
		f0 += 0.5;
	}
	f0 = f0 * HTotal;
	f0 = f0 * Vert001Hz;
	f0 = f0 / 100.0;

	return (long)f0;
}

long CalcHorizontalHz(long dotclockHz, int HTotal)
{
	return dotclockHz / HTotal;
}

long CalcVert001Hz(long dotclockHz, int HTotal, int VTotal, int interlace)
{
	double f0;

	f0 = (double)VTotal;
	if(interlace != 0){
		f0 += 0.5;
	}
	f0 = f0 * HTotal;
	f0 = dotclockHz / f0;
	f0 = f0 * 100.0;

	return (long)f0;
}



void DispCrtcParam(CRTCPARAM *cp)
{
	logerror("DotClock..... %.4lf MHz\n",
						(double)(cp->dotclockHz / 1000.0) / 1000.0);
	logerror("HTotal....... %d\n", cp->HTotal);
	logerror("HDispEnd..... %d\n", cp->HDisp);
	logerror("HBlankStart.. %d\n", cp->HBStart);
	logerror("HSyncStart... %d\n", cp->HSStart);
	logerror("HSyncEnd..... %d\n", cp->HSEnd);
	logerror("HBlankEnd.... %d\n", cp->HBEnd);
	logerror("VTotal....... %d\n", cp->VTotal);
	logerror("VDispEnd..... %d\n", cp->VDisp);
	logerror("VBlankStart.. %d\n", cp->VBStart);
	logerror("VSyncStart... %d\n", cp->VSStart);
	logerror("VSyncEnd..... %d\n", cp->VSEnd);
	logerror("VBlankEnd.... %d\n", cp->VBEnd);
	logerror("Horz......... %.2lf KHz\n",
						(double)cp->horzHz / 1000.0);
	logerror("Vert......... %.2lf Hz\n",
						(double)cp->vert001Hz / 100.0);

	if(cp->hpolarity == 0){
		logerror("HorzSync..... Positive (Low  -> High)\n");
	}else{
		logerror("HorzSync..... Negative (High -> Low)\n");
	}
	if(cp->vpolarity == 0){
		logerror("VertSync..... Positive (Low  -> High)\n");
	}else{
		logerror("VertSync..... Negative (High -> Low)\n");
	}

	if(cp->interlace == 0){
		logerror("non interlaced mode\n");
	}else{
		logerror("interlaced mode\n");
	}
	if(cp->doublescan != 0){
		logerror("double scan\n");
	}

}


void SetHVPolarity(int HPolar, int VPolar)
{
	BYTE d0;

	if(HPolar != 0) HPolar = 1;
	if(VPolar != 0) VPolar = 1;

	d0 = _in(0x3CC);
	d0 = d0 & 0x3F;
	d0 = d0 | (VPolar << 7) | (HPolar << 6);
	_out(0x3C2, d0);
}

void SetTextHeight(int height)
{
	BYTE d0;

	height = (height - 1) & 0x1F;
	d0	= GetCrtcReg(0x09) & ~0x1F;
	d0 |= height;
	SetCrtcReg(0x09, d0);
}





DWORD BitMov(DWORD dstdata, int dstbitpos, DWORD srcdata, int srcbitpos)
{
	DWORD d0;
	DWORD d1;

	d1 = srcdata >> srcbitpos;
	d1 = d1 & 1;
	d0 = dstdata & ~(1 << dstbitpos);
	d1 = d1 << dstbitpos;
	d0 = d0 | d1;
	return d0;
}



void SetSeqReg(BYTE __idx, BYTE data)
{
	_out(0x3c4, __idx);
	_out(0x3c5, data);
}

BYTE GetSeqReg(BYTE __idx)
{
	BYTE d0;

	_out(0x3c4, __idx);
	d0 = _in(0x3c5);
	return d0;
}


BYTE GetAtrbData(BYTE __idx)
{
	int d0;

	if((_inb(0x3CC) & 0x01) == 0)
		d0 = 0x3BA;
	else
		d0 = 0x3DA;

	_in(d0);
	_out(0x3c0, __idx | (1 << 5));
	return _in(0x3c1);
}

void SetAtrbData(BYTE __idx, BYTE data)
{
	int d0;

	if((_inb(0x3CC) & 0x01) == 0)
		d0 = 0x3BA;
	else
		d0 = 0x3DA;

	_in(d0);
	_out(0x3c0, __idx | 0x20);
	_out(0x3c0, data);
}


int GetCrtcAddress(void)
{
	if((_inb(0x3CC) & 0x01) == 0)
		return 0x3B4;
	else
		return 0x3D4;
}

void SetCrtcReg(BYTE __idx, BYTE data)
{
	_outb(GetCrtcAddress(),    __idx);
	_outb(GetCrtcAddress() + 1, data);
}

BYTE GetCrtcReg(BYTE __idx)
{
	BYTE d0;

	_outb(GetCrtcAddress(), __idx);
	d0 = _inb(GetCrtcAddress() + 1);
	return d0;
}


void SetGfxcReg(BYTE __idx, BYTE data)
{
	_outb(0x3ce, __idx);
	_outb(0x3cf, data);
}

BYTE GetGfxcReg(BYTE __idx)
{
	BYTE d0;

	_outb(0x3ce, __idx);
	d0 = _inb(0x3cf);
	return d0;
}



int GetSeqDivider(void)
{
	int d0;

	d0	= (GetSeqReg(0x01) & 0x08) >>3;
	d0	= d0 + 1;
	return d0;
}

void SetSeqDivider(int d0)
{
	int d1;

	d0 -= 1;
	if(d0 != 0) d0 = 1 << 3;
	d1 = GetSeqReg(0x01) & ~0x08;
	d1 = d1 | d0;
	SetSeqReg(0x01, d1);
}


void _out(int edx, BYTE al)
{
	outportb(edx, al);
}
BYTE _in(int edx)
{
	return inportb(edx);
}

void _outb(int edx, BYTE al)
{
	outportb(edx, al);
}
BYTE _inb(int edx)
{
	return inportb(edx);
}

void _outw(int edx, WORD ax)
{
	outportw(edx, ax);
}
WORD _inw(int edx)
{
	return inportw(edx);
}

void _outl(int edx, DWORD eax)
{
	outportl(edx, eax);
}
DWORD _inl(int edx)
{
	return inportl(edx);
}


void DisableGenerateSignal(void)
{
	BYTE d0;

	d0 = GetCrtcReg(0x17);
	d0 = d0 & 0x7F;
	SetCrtcReg(0x17, d0);
}

void EnableGenerateSignal(void)
{
	BYTE d0;

	d0 = GetCrtcReg(0x17);
	d0 = d0 | 0x80;
	SetCrtcReg(0x17, d0);
}

void DisableStretch(void)
{
	stretch = 0;
	scanlines = 0;
}





static void ChangeSyncWidth(CRTCPARAM *cp, int dx, int dy)
{
	if(dx < 0){
		if((cp->HSEnd + dx) > cp->HSStart){
			cp->HSEnd	+= dx;
		}
	}else{
		if((cp->HSEnd + dx) < (cp->HTotal - 8)){
			cp->HSEnd	+= dx;
		}
	}

	if(dy < 0){
		if((cp->VSEnd + dy) > cp->VSStart){
			cp->VSEnd	+= dy;
		}
	}else{
		if((cp->VSEnd + dy) < (cp->VTotal - 1)){
			cp->VSEnd	+= dy;
		}
	}
}


void RegCRTCPARAM(CRTCPARAM *cp)
{
	rcp = *cp;
}

void GetCRTCPARAM(CRTCPARAM *cp)
{
	*cp = rcp;
}



static void change_param(CRTCPARAM *cp)
{
	if(SVGA15KHzdriver){
		if(SVGA15KHzdriver->centerSVGA15KHzmode != NULL){
			RegCRTCPARAM(cp);
			SVGA15KHzdriver->centerSVGA15KHzmode(0, 0);
		}
	}
}


void SetCRTCPARAM_HSyncPos(CRTCPARAM *cp, int dx)
{
	Centering(cp, dx, 0);
	change_param(cp);
}

void SetCRTCPARAM_HSyncWidth(CRTCPARAM *cp,int dx)
{
	ChangeSyncWidth(cp, dx, 0);
	change_param(cp);
}

void SetCRTCPARAM_VSyncPos(CRTCPARAM *cp, int dy)
{
	Centering(cp, 0, dy);
	change_param(cp);
}

void SetCRTCPARAM_VSyncWidth(CRTCPARAM *cp, int dy)
{
	ChangeSyncWidth(cp, 0, dy);
	change_param(cp);
}

void SetCRTCPARAM_Apply(CRTCPARAM *cp)
{
	int d0;

	change_param(cp);

	if(use_crtc_data_base != 0){
		/* save crtcparam to crtc data base */
		if((video_orientation & ORIENTATION_SWAP_XY) == 0)
			d0 = 0; 	// horz
		else
			d0 = 1; 	// vert
		if(use_y_stretch != 0) d0 += 2;
		set_crtc_param_cfg((char*)Machine->gamedrv->name, svga15kdrv_monitorname,
								&orientation_string[d0][0], cp);
	}
}

void SetCRTCPARAM_VSyncAdjust(CRTCPARAM *cp, int dx)
{
	if(dx < 0){
		if((cp->interlaceadjust + dx) >= 0)
			cp->interlaceadjust += dx;
		else
			cp->interlaceadjust = 0;
	}else{
		if((cp->interlaceadjust + dx) <= cp->HTotal)
			cp->interlaceadjust += dx;
		else
			cp->interlaceadjust = cp->HTotal;
	}

	change_param(cp);
}
