#include "mamalleg.h"
#include "driver.h"
#include <dos.h>
#include <conio.h>
#include <stdlib.h>
#include "ticker.h"
#include "wss.h"
#include "../sound/ratecnv.h"


#define FRACTIONBITS	12
#define FRACTIONMASK	((1 << FRACTIONBITS) - 1)

static INT16 *conversion_buffer = NULL;
static INT16 *result_buffer 	= NULL;
static int conversion_buffer_length = 0;
static int result_buffer_length = 0;

static int default_device_samples_per_frame = 800;

/* dummy param for config.c */
int sampleratedetect = 1;

/* request frame skip flag */
int request_frame_skip = 0;

static int num_used_opl;
int nominal_sample_rate;
int soundcard, usestereo;
int attenuation = 0;
static int master_volume = 256;

static int stream_playing;
static int stream_cache_stereo;
static INT16 *stream_cache_data;
static int stream_cache_len = 0;
static int stream_cache_valid = 0;

static int srcnv_channel[2];

extern float fps_value;

int reverse_stereo = 0;
int sb_cursor_offset = 0;
int oplbaseindex = 0;
int oplbaseio = 0x388;
int rec_wav = 0;
int device_sample_rate = 0;
int internal_sample_rate = 0;

static float sound_watermark = 2.2;

static FILE* fpwav = NULL;
static INT32 wav_bytes = 0;
static INT8 riffheader[44];

#define DchunkID				 0
#define DchunkSize				 4
#define DformType				 8
#define Dfmt_chunkID			12
#define Dfmt_chunkSize			16
#define Wfmt_waveFormatType 	20
#define Wfmt_channel			22
#define Dfmt_samplesPerSec		24
#define Dfmt_bytesPerSec		28
#define Wfmt_blockSize			32
#define Wfmt_bitsPerSample		34
#define Ddata_chunkID			36
#define Ddata_chunkSize 		40
#define Ddata_PCM				44


static FILE* init_wav_header(FILE* fp, int stereo, int sample_rate)
{
	INT16 d0;
	size_t s0;

	if(fp == NULL) return NULL;

	riffheader[DchunkID + 0] = 'R';
	riffheader[DchunkID + 1] = 'I';
	riffheader[DchunkID + 2] = 'F';
	riffheader[DchunkID + 3] = 'F';
	*(INT32*)&riffheader[DchunkSize] = 36;
	riffheader[DformType + 0] = 'W';
	riffheader[DformType + 1] = 'A';
	riffheader[DformType + 2] = 'V';
	riffheader[DformType + 3] = 'E';

	riffheader[Dfmt_chunkID + 0] = 'f';
	riffheader[Dfmt_chunkID + 1] = 'm';
	riffheader[Dfmt_chunkID + 2] = 't';
	riffheader[Dfmt_chunkID + 3] = ' ';
	*(INT32*)&riffheader[Dfmt_chunkSize] = 16;
	*(INT16*)&riffheader[Wfmt_waveFormatType] = 1;	  /* linear PCM */
	if(stereo != 0){
		d0 = 2;
	}else{
		d0 = 1;
	}
	*(INT16*)&riffheader[Wfmt_channel] = d0;
	*(INT32*)&riffheader[Dfmt_samplesPerSec] = sample_rate;
	*(INT32*)&riffheader[Dfmt_bytesPerSec] = sample_rate * (16/8) * d0;
	*(INT16*)&riffheader[Wfmt_blockSize] = (16/8) * d0;
	*(INT16*)&riffheader[Wfmt_bitsPerSample] = 16;

	riffheader[Ddata_chunkID + 0] = 'd';
	riffheader[Ddata_chunkID + 1] = 'a';
	riffheader[Ddata_chunkID + 2] = 't';
	riffheader[Ddata_chunkID + 3] = 'a';
	*(INT32*)&riffheader[Ddata_chunkSize] = 0;

	s0 = fwrite(riffheader, 1, 44, fp);
	if(s0 != 44){
		fclose(fp);
		return NULL;
	}
	return fp;
}

static void exit_wav_header(FILE* fp)
{
	if(fp == NULL) return;

	*(INT32*)&riffheader[DchunkSize] = 36 + wav_bytes;
	*(INT32*)&riffheader[Ddata_chunkSize] = wav_bytes;

	fseek(fp, 0, SEEK_SET);
	fwrite(riffheader, 1, 44, fp);
	fclose(fp);
}


static FILE* write_wave_data(FILE* fp, INT16 *data, INT32 len, int stereo)
{
	size_t s0;
	int d0;

	d0 = 1;
	if(stereo != 0) d0 = 2;
	wav_bytes += 2 * d0 * len;
	s0 = fwrite(data, 1, 2 * d0 * len, fp);
	if(s0 != (2 * d0 * len)){
		fclose(fp);
		return NULL;
	}
	return fp;
}


int is_requested_frame_skip_by_sound(void)
{
	int d0;

	d0 = request_frame_skip;
	request_frame_skip = 0;
	return d0;
}


static float frames_per_second(void)
{
	float rate;

	if(fps_value == 0.0){
		rate = Machine->drv->frames_per_second;
	}else{
		rate = fps_value;
	}
	return rate;
}


void set_opl_base(void)
{
	switch(oplbaseindex){
		case 0:
			oplbaseio = 0x388;
			break;
		case 1:
			oplbaseio = 0x398;
			break;
		case 2:
			oplbaseio = 0x220;
			break;
		case 3:
			oplbaseio = 0x230;
			break;
		case 4:
			oplbaseio = 0x240;
			break;
		case 5:
			oplbaseio = 0x250;
			break;
		case 6:
			oplbaseio = 0x260;
			break;

		default:
			oplbaseio = 0x388;
	}
	logerror("OPLbase 0x%x\n", oplbaseio);
}



int msdos_init_seal (void)
{
	return 0;
}




int msdos_init_sound(void)
{
	int i;
	int rate;
	float samples_per_frame;

	w_set_sb_cursor_offset(sb_cursor_offset);

	if (soundcard == -1)
	{

		printf("\nSelect the audio device:\n");
		printf("  0. Eternal Silence\n");
		printf("  1. Sound Blaster\n");
		printf("  4. Ultrasound Max\n");
		printf("  5. Ultrasound\n");
		printf("  6. Windows Sound System\n");
		printf("  7. Ensoniq Soundscape\n");
		printf("  8. ESS Audiodrive\n");
		printf("  9. Sound Blaster Pro (for WDM SBPro emulation, RDTSC reqd.)\n");
		printf("\n");

		i = getch();
		soundcard = i - '0';
		if ( !((0 <= soundcard) && (soundcard <= 9)) )
		{
			printf("audio initialization failed\n");
			soundcard = -1;
			return 1;
		}

	}

	if(soundcard != 0)
		set_opl_base();

	if(soundcard == 0)
		if(device_sample_rate == 0){
			rate = (int)((frames_per_second() * 10) + 0.5);
		}else{
			rate = device_sample_rate;
		}
	else
		rate = Machine->sample_rate;

	if( w_sound_device_init(soundcard, rate) == _FALSE ){
		printf("%s\n", w_get_error_message() );
		printf("audio initialization failed\n");
		soundcard = -1;
		return 1;
	}
	if(reverse_stereo) w_reverse_stereo(_TRUE);

	logerror("%s%7d Hz\n", w_get_device_name(), (int)w_get_nominal_sample_rate());

	/* update the Machine structure to reflect the actual sample rate */
	osd_set_mastervolume(attenuation);	/* set the startup volume */

	w_set_watermark(2.2, w_get_nominal_sample_rate()/60);

	if(device_sample_rate == 0){
		device_sample_rate = w_get_actual_sample_rate();
		logerror("device sample rate: %d\n", device_sample_rate);
		if(device_sample_rate == 0){
			printf("device not found\n");
			printf("audio initialization failed\n");
			soundcard = -1;
			return 1;
		}
	}else{
		logerror("specified device sample rate: %d\n", device_sample_rate);
	}

	if(internal_sample_rate == 0){
		logerror("internal_sample_rate == 0\n");
		logerror("Machine->sample_rate = device_sample_rate\n");;
		Machine->sample_rate = device_sample_rate;
	}else{
		logerror("internal_sample_rate != 0\n");
		conversion_buffer_length = internal_sample_rate / 4;
		result_buffer_length	 = (device_sample_rate / 4) * 2;	// for stereo
		conversion_buffer = (INT16*)malloc(conversion_buffer_length * sizeof(INT16));
		result_buffer	  = (INT16*)malloc(result_buffer_length 	* sizeof(INT16));
		if(conversion_buffer == NULL || result_buffer == NULL){
			logerror("failed buffer allocation\n");
			free(conversion_buffer);
			free(result_buffer);
			return 1;
		}
		logerror("conversion buffer size %d\n", conversion_buffer_length);
		logerror("result buffer size %d\n", result_buffer_length);
		Machine->sample_rate = internal_sample_rate;
		srcnv_init(1);		/* use lpf */
		srcnv_channel[0] = srcnv_new(internal_sample_rate, device_sample_rate);
		srcnv_channel[1] = srcnv_new(internal_sample_rate, device_sample_rate);
	}
	logerror("Machine->sample_rate: %d\n", Machine->sample_rate);

	if(sound_watermark == 0.0) sound_watermark = 2.2;
	samples_per_frame = device_sample_rate/frames_per_second();
	logerror("device sample rate %d Hz, %.2f FPS\n", device_sample_rate, frames_per_second());
	logerror("samples per frame %.2f\n", samples_per_frame);
	logerror("watermark %.2f * %.2f = %d samples\n", sound_watermark, samples_per_frame,
			(int)(sound_watermark * samples_per_frame));

	default_device_samples_per_frame = samples_per_frame;
	w_set_watermark(sound_watermark, samples_per_frame);

	stream_cache_valid = 0;
	num_used_opl = 0;

	if(rec_wav != 0){
		fpwav = fopen("sound.wav", "wb+");
	}

	return 0;
}






void msdos_shutdown_sound(void)
{
	int chip,n;

	if (Machine->sample_rate != 0)
	{
		w_sound_device_exit();
	}

	for(chip=0;chip<num_used_opl;chip++)
	{
		/* silence the OPL */
		for (n = 0x40;n <= 0x55;n++)
		{
			osd_opl_control(chip,n);
			osd_opl_write(chip,0x3f);
		}
		for (n = 0x60;n <= 0x95;n++)
		{
			osd_opl_control(chip,n);
			osd_opl_write(chip,0xff);
		}
		for (n = 0xa0;n <= 0xb0;n++)
		{
			osd_opl_control(chip,n);
			osd_opl_write(chip,0);
		}
	}

	if(fpwav != NULL){
		exit_wav_header(fpwav);
		fpwav = NULL;
	}

	srcnv_exit();

	free(conversion_buffer);
	free(result_buffer);
}







/* global sample tracking */
static double samples_per_frame;
static double samples_left_over;
static UINT32 samples_this_frame;


int osd_start_audio_stream(int stereo)
{
	if (Machine->sample_rate == 0) return 0;

	if (stereo) stereo = 1; /* make sure it's either 0 or 1 */

	stream_cache_stereo = stereo;

	/*	recording  */
	if(fpwav != NULL){
		fpwav = init_wav_header(fpwav, stereo, Machine->sample_rate);
	}

	/* determine the number of samples per frame */
	samples_per_frame = (double)Machine->sample_rate / frames_per_second();

	/* compute how many samples to generate this frame */
	samples_left_over = samples_per_frame;
	samples_this_frame = (UINT32)samples_left_over;
	samples_left_over -= (double)samples_this_frame;

	stream_playing = 1;
	stream_cache_valid = 0;

	return samples_this_frame;
}




void osd_stop_audio_stream(void)
{
	w_clear_buffer();
	stream_cache_valid = 0;
}



static void copy_stereo_to_mono(INT16 *src, INT16 *dst, int length)
{
	int d0;

	d0 = 0;
	while(d0 < length){
		dst[d0] = src[d0*2];
		d0 += 1;
	}
}

static int rate_conversion_and_write_to_buffer(INT16 *data, int length, int stereo)
{
/* return length of converted data */

	INT16 *outbuff;
	int outlen;
	int result_buff_index;
	int channel;
	int d0;

	if(conversion_buffer == NULL || result_buffer == NULL) return 0;
	if(length > conversion_buffer_length) return 0;

	if(stereo){
		/* stereo */
		channel = 0;
		while(channel < 2){
			copy_stereo_to_mono(data + channel, conversion_buffer, length); 		// left then right
			srcnv_apply_init(srcnv_channel[channel], conversion_buffer, length);	// conversion start
			result_buff_index = 0;
			while(1){
				srcnv_apply(srcnv_channel[channel]);
				outbuff = srcnv_get_outbuffer_address(srcnv_channel[channel]);
				outlen	= srcnv_get_outbuffer_length(srcnv_channel[channel]);
				d0 = 0;
				while(d0 < outlen){
					if(result_buff_index >= result_buffer_length/2) break;
					/*	leftdata 0, 2, 4...  rightdata 1, 3, 5...  */
					result_buffer[channel + (result_buff_index * 2)] = outbuff[d0];
					result_buff_index += 1;
					d0 += 1;
				}
				if( srcnv_is_finished(srcnv_channel[channel]) != 0 ) break;
			}
			srcnv_apply_exit(srcnv_channel[channel]);
			channel += 1;
		}
	}else{
		/* mono */
		channel = 0;
		srcnv_apply_init(srcnv_channel[channel], data, length); 	// conversion start
		result_buff_index = 0;
		while(1){
			srcnv_apply(srcnv_channel[channel]);
			outbuff = srcnv_get_outbuffer_address(srcnv_channel[channel]);
			outlen	= srcnv_get_outbuffer_length(srcnv_channel[channel]);
			d0 = 0;
			while(d0 < outlen){
				if(result_buff_index >= result_buffer_length) break;
				result_buffer[result_buff_index++] = outbuff[d0];
				d0 += 1;
			}
			if( srcnv_is_finished(srcnv_channel[channel]) != 0 ) break;
		}
		srcnv_apply_exit(srcnv_channel[channel]);
	}

	return result_buff_index;
}


static void updateaudiostream(void)
{
	/* from video.c */
	extern int get_frame_skip_level(void);
	extern int is_vsync_wait_enabled(void);
	extern int throttle;
	INT16 *data = stream_cache_data;
	int stereo = stream_cache_stereo;
	int len = stream_cache_len;
	int d0;
	int frame_skip_level;
	TICKER prev;

	if(data == 0) stream_cache_valid = 0;
	if(stream_cache_valid != 0){
		if(len > 0){
			if(fpwav != NULL){
				fpwav = write_wave_data(fpwav, data, len, stereo);
			}
			if(internal_sample_rate != 0){
				/* we need rate conversion */
				len  = rate_conversion_and_write_to_buffer(data, len, stereo);
				data = result_buffer;
			}
		}
	}

	frame_skip_level = get_frame_skip_level();
	if (throttle)	/* sync with audio only when speed throttling is not turned off */
	{
		profiler_mark(PROFILER_IDLE);
		prev = ticker();
		while(1){
			/*	2:full, 1:good, 0:ok, -1:lack, -2:overrun  */
			d0 = w_get_watermark_status();
			if(frame_skip_level <= 1){									// 0, 1
				if(d0 == -1 || d0 == -2)
					request_frame_skip = 1;
				else
					request_frame_skip = 0;
				if(d0 == 0 || d0 == -1) break;
			}else if(frame_skip_level == 2){							// 2
				if(d0 == -1 || d0 == -2)
					request_frame_skip = 1;
				else
					request_frame_skip = 0;
				if(d0 == 1 || d0 == 0 || d0 == -1) break;
			}else{														// 3, 4...
				if(d0 == 0 || d0 == -1 || d0 == -2)
					request_frame_skip = 1;
				else
					request_frame_skip = 0;
				if(d0 == 1 || d0 == 0 || d0 == -1) break;
			}
			if(d0 == -2){
				w_reset_watermark_ex(default_device_samples_per_frame);
				break;
			}
			if(is_vsync_wait_enabled() != 0) break; 	/* pass through if vsync enabled */
			if( (ticker() - prev) > TICKS_PER_SEC/2 ) break;
		}
		profiler_mark(PROFILER_END);
	}

	if(stream_cache_valid != 0){
		if(len > 0){
			w_lock_mixing_buffer(len);
			if(stereo)
				w_mixing_stereo(data, len, 256, 256);
			else
				w_mixing(data, len, 256, 256);
			w_unlock_mixing_buffer();
		}
	}else{
		w_lock_mixing_buffer(default_device_samples_per_frame);
		w_mixing_zero();
		w_unlock_mixing_buffer();
	}

	stream_cache_valid = 0;
}






int osd_update_audio_stream(INT16 *buffer)
{
	extern int video_sync;

	stream_cache_data = buffer;
	stream_cache_len  = samples_this_frame;

	if(!video_sync){
		/* compute how many samples to generate next frame */
	   samples_left_over += samples_per_frame;
	   samples_this_frame = (UINT32)samples_left_over;
	   samples_left_over -= (double)samples_this_frame;
	}else{
		samples_this_frame = srcnv_get_src_sample_length(
					Machine->sample_rate, device_sample_rate, w_get_next_req());
		samples_this_frame += 1;
	}

	stream_cache_valid = 1;

	return samples_this_frame;
}

int msdos_update_audio(void)
{
	if (Machine->sample_rate == 0) return 0;

	profiler_mark(PROFILER_MIXER);
	updateaudiostream();
	profiler_mark(PROFILER_END);
	return 1;
}

/* attenuation in dB */
void osd_set_mastervolume(int _attenuation)
{
	float volume;

	if (_attenuation > 0) _attenuation = 0;
	if (_attenuation < -32) _attenuation = -32;

	attenuation = _attenuation;

	volume = 256.0; /* range is 0-256 */
	while (_attenuation++ < 0)
		volume /= 1.122018454;	/* = (10 ^ (1/20)) = 1dB */


	master_volume = volume;

	w_set_master_volume(master_volume);
}


int osd_get_mastervolume(void)
{
	return attenuation;
}

void osd_sound_enable(int enable_it)
{
	if (enable_it)
	{
		w_set_master_volume(master_volume);
	}
	else
	{
		w_set_master_volume(0);
		w_clear_buffer();
	}
	stream_cache_valid = 0;
}



/* linux sound driver opl3.c does a so called tenmicrosec() delay */
static void tenmicrosec(void)
{
	int i;
	for (i = 0; i < 16; i++)
		inportb(0x80);
}

//#define MAX_OPLCHIP 2  /* SOUND BLASTER 16 or compatible ?? */
#define MAX_OPLCHIP 1  /* SOUND BLASTER pro compatible ??  */

void osd_opl_control(int chip,int reg)
{
	if (Machine->sample_rate == 0) return;
	if (soundcard == 0) return;

	if (chip >= MAX_OPLCHIP ) return;
	tenmicrosec();
	outportb(oplbaseio + (chip*2), reg);
}

void osd_opl_write(int chip,int data)
{
	if (Machine->sample_rate == 0) return;
	if (soundcard == 0) return;

	if (chip >=MAX_OPLCHIP ) return;
	tenmicrosec();
	outportb( (oplbaseio+1) + (chip*2), data);

	if(chip >= num_used_opl) num_used_opl = chip+1;
}
