
int srcnv_init(int use_filter_flag);
/*
  use_filter_flag != 0, using lpf (slow)
  use_filter_flag == 0, no filter (fast)

  normally set 1 to use_filter_flag, but 0 for slow cpu, a x86 cpu have no mmx insn
  to avoid invalid instruction exception, etc.
*/

void srcnv_exit(void);
int srcnv_new(double srcfreq, double dstfreq);
int srcnv_change(int srcnv_channel, double srcfreq, double dstfreq);


int srcnv_apply(int srcnv_channel);
int srcnv_apply_init(int srcnv_channel, INT16* src, int length);
int srcnv_apply_exit(int srcnv_channel);
INT16* srcnv_get_outbuffer_address(int srcnv_channel);
int srcnv_get_outbuffer_length(int srcnv_channel);
int srcnv_is_finished(int srcnv_channel);

int srcnv_get_filter_taps(void);
int srcnv_get_src_sample_length(double srcfreq, double dstfreq, int dst_length);

int lpf_new(double fs, double fc);
int lpf_change(int lpf_channel, double fs, double fc);
void lpf_apply(int lpf_channel, INT16* src, int length);


/********************************************************************

*How to use LPF*

#define SRC_FREQ        22050
#define CUTOFF_FREQ      5000
#define BUFFLENGTH  1024

{
    INT16 buff[BUFFLENGTH];
    int d0;
    int loop;
    int c = 0;
    int delta = 0x0400;
    int lpf_channel;

    srcnv_init(1);      // use lpf
    lpf_channel = lpf_new(SRC_FREQ, CUTOFF_FREQ);
    loop = 0;
    while(loop < SEVERAL_TIMES){
        d0 = 0;
        while(d0 < BUFFLENGTH){
            buff[d0] = (c & 0x2000) - 0x1000;   // generate square wave
            c += delta;
            d0 += 1;
        }
        lpf_apply(lpf_channel, buff, BUFFLENGTH);
//        PLAY_PCM(SRC_FREQ, buff, BUFFLENGTH);
        loop += 1;
    }
    srcnv_exit();
}



*How to use rate converter*

#define SRC_FREQ    22050
#define DST_FREQ    48000
#define BUFFLENGTH  1024

{
    INT16 buff[BUFFLENGTH];
    INT16 dstbuff[BUFFLENGTH * 16];
    int dstbuff_index;
    INT16 *outbuff;
    int outlen;
    int d0;
    int loop;
    int c = 0;
    int delta = 0x0400;
    int srcnv_channel;
    int samples_converted;

    srcnv_init(1);      // use lpf
    srcnv_channel = srcnv_new(SRC_FREQ, DST_FREQ);
    loop = 0;
    while(loop < SEVERAL_TIMES){
        d0 = 0;
        while(d0 < BUFFLENGTH){
            buff[d0] = (c & 0x2000) - 0x1000;   // generate square wave
            c += delta;
            d0 += 1;
        }
        samples_converted = 0;
        srcnv_apply_init(srcnv_channel, buff, BUFFLENGTH);      // conversion start
        while(1){
            srcnv_apply(srcnv_channel);
            outbuff = srcnv_get_outbuffer_address(srcnv_channel);
            outlen  = srcnv_get_outbuffer_length(srcnv_channel);
            d0 = 0;
            while(d0 < outlen){
                dstbuff[samples_converted++] = outbuff[d0];
                d0 += 1;
            }
            if( srcnv_is_finished(srcnv_channel) != 0 ) break;
        }
        srcnv_apply_exit(srcnv_channel);
//        PLAY_PCM(DST_FREQ, dstbuff, samples_converted);
        loop += 1;
    }
    srcnv_exit();
}




********************************************************************/
